"""Main file to run functions in the function file."""

# %% IMPORT PACKAGES

import os
import pandas as pd
import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.ticker as plticker
from matplotlib.cm import ScalarMappable
import seaborn as sns
sns.set_theme(palette="colorblind", style="whitegrid")

from classes.database import Database
from classes.dataset import Dataset
from classes.timeseries import Timeseries
from classes.plot import FigurePlot
from classes.error import ErrorMark
import functions.functions as fn
import functions.hydroyear as hy

# %% LOAD DATA
# %%% Load data
db = fn.load_database(path="..\\data\\")

raise ErrorMark("Stop after loading data.")

# %% FIGURE 4: DESCRIPTION CLIMATE
data = fn.plot_climate(db)  # Data not included but can be downloaded from:
                            # https://www.knmi.nl/nederland-nu/klimatologie/uurgegevens

# %%% Study period
data = fn.plot_climate_studyperiod(db)  # Data not included but can be downloaded from:
                                        # https://www.knmi.nl/nederland-nu/klimatologie/uurgegevens

# %% FIGURE 5: ANNUAL WATER BALANCE
data = fn.plot_Q_annual(db)

# %%% Difference tb and pump on annual level
ds = Dataset(db)
index = hy.get_index_hydroyear()
ds.add_timeseries({'dtype': 'Q', 'layer': 'CL', 'device': 'tb'})
ds.add_timeseries({'dtype': 'Q', 'layer': 'CL', 'device': 'p'})
ds.add_timeseries({'dtype': 'Q', 'layer': 'CB', 'device': 'tb'})
ds.add_timeseries({'dtype': 'Q', 'layer': 'CB', 'device': 'p'})
ds.set_properties_timeseries(resample={'scale': '1Y', 'method': 'sum', 'index': index},
                             period=['2010-04-01', '2023-03-31'])
data = ds.get_dataframe()

# %%% Change in storage each year
index = hy.get_index_hydroyear()
db.calculate_avg(name='sm__avg_TS', ts_dic={'dtype': 'sm', 'layer': ['5cm', '15cm']},
                 resample={'method': 'first', 'scale': '1Y', 'index': index})
db.calculate_avg(name='sm__avg_SS', ts_dic={'dtype': 'sm', 'layer': ['40cm', '80cm']},
                 resample={'method': 'first', 'scale': '1Y', 'index': index})
ds = Dataset(db)
index = hy.get_index_hydroyear()
ds.add_timeseries({'dtype': 'sm', 'device': 'avg'})
data = ds.get_dataframe()

# %% FIGURE 6: DAILY SCALE
data = fn.plot_Q_daily(db)

# %% FIGURE 7: CORRELATION FLOWS WITH YEAR AS COLOR
data = fn.plot_Q_daily_corr_year(db)

# %%% Each year in a separate subplot
data = fn.plot_Q_daily_corr_year_subplots(db)

# %% FIGURE 10: Plot water contents
data = fn.plot_watercontent(db)

# %% FIGURE 11: Frequency distribution of flows
data = fn.plot_frequency_distribution_Q_daily(db)

# %%% Subset between 0 and 1 mm/d
data = fn.plot_frequency_distribution_Q_daily_subset(db)

# %% EXPORT CURRENT FIGURE

# fmt = 'eps'
fmt = 'jpeg'
cwd = os.getcwd()
os.chdir("C:\\Users\\vdbri\\Documents\\TUDELFT\\Afstuderen\\5_Paper\\0_figs")
# plt.savefig("FIG4_DescriptionClimate.eps", format='eps')
# plt.savefig("FIG5_AnnualWaterBalance.eps", format='eps')
# plt.savefig("FIG6_DailyWaterBalance.eps", format='eps')
# plt.savefig("FIG7_CorrelationFlowsPerYear."+fmt, format=fmt)
plt.savefig("FIG7_CorrelationFlowsPerYear.png")
os.chdir(cwd)

# %% EXPORT DATA
path = "C:\\Users\\vdbri\\Documents\\TUDELFT\\Afstuderen\\0_Data\\Export\\"

# %%% Outflow
ds = Dataset(db)
ds.add_timeseries(ts_dic={'dtype': 'Q', 'layer': 'CL', 'device': 'tb'})
ds.add_timeseries(ts_dic={'dtype': 'Q', 'layer': 'CB', 'device': 'tb'})
ds.add_timeseries(ts_dic={'dtype': 'Q', 'layer': 'CL', 'device': 'p'})
ds.add_timeseries(ts_dic={'dtype': 'Q', 'layer': 'CB', 'device': 'p'})
data = ds.get_dataframe()
data.columns = ['Outflow capillary layer tipping bucket (mm)',
                'Outflow capillary block tipping bucket (mm)',
                'Outflow capillary layer pump (mm)',
                'Outflow capillary block pump (mm)']
data.to_csv(path+"Outflow.csv")

# %%% Water content
ds = Dataset(db)
ds.add_timeseries(ts_dic={'dtype': 'sm'})
ds.set_properties_timeseries(resample={'method': 'avg', 'scale': '1H'})
data = ds.get_dataframe()
data.columns = ['Soil moisture ' +
                ['upslope ', 'midslope ', 'downslope '][int(ii/4)] +
                name.split('_')[-1] +
                ' depth (vol.%)'
                for ii, name in enumerate(data.columns)]
data.to_csv(path+"WaterContent.csv")
