import numpy as np
import pandas as pd

from classes.timeseries import Timeseries
from classes.dataset import Dataset
from classes.error import ErrorMark


class Database:

    def __init__(self):
        self.timeseries = []  # list to store timeseries objects
        self.names_ts = []  # list to store the names of the timeseries

    def add_timeseries(self, ts_new):
        """Add timeseries to list of timeseries in this dataset."""
        # If name already exists in dataset, change data
        if ts_new.name in self.names_ts:
            ts_old = self.timeseries[self.names_ts.index(ts_new.name)]
            ts_old.data = ts_new.data
        # If not, add timeseries to dataset
        else:
            self.timeseries.append(ts_new)
            self.names_ts.append(ts_new.name)

    def get_timeseries(self, name_dic):
        """Return list of timeseries objects for given dtype, dbase, device, layer or state."""
        # Extract given dictionary: put given conditions in right order
        var = [[], [], [], [], [], []]
        var_str = ['name', 'dtype', 'dbase', 'device', 'layer', 'state']
        for i, v in enumerate(var_str):
            if v in name_dic.keys():
                var[var_str.index(v)] = name_dic[v]
        # List timeseries in dataset who comply with given search terms.
        ts_match = []
        for ts in self.timeseries:
            if ts.check_similarity(*var):
                ts_match.append(ts)
        if len(ts_match) == 0:
            raise ErrorMark(f"Given dict {name_dic} has no match.")
        elif len(ts_match) == 1:
            ts_match = ts_match[0]
        # Return single object if only one timeseries matches, else return complete list.
        return ts_match

    def calculate_discrete_timeseries(self, ts_dic):
        """Return a dataset with discrete values of timeseries of input dataset."""
        ts_list = self.get_timeseries(ts_dic)
        ts_list = ts_list if isinstance(ts_list, list) else [ts_list]
        # Generate new Timeseries objects for each Timeseries object in ts_list
        for ts in ts_list:
            ts_new = Timeseries(ts.name[:-3]+'dis', ts.data)
            # Calculate discrete data for this new Timeseries object
            ts_new.cumulative_to_discrete_data()
            self.add_timeseries(ts_new)

    def calculate_discharge_timeseries(self, area=500):
        """Build discharge timeseries from switches and volume datasets."""
        # Names of Timeseries objects to build
        names_list = ['Q_cars_tb_CL_dis', 'Q_cars_tb_CB_dis', 'Q_cars_tb_DM_dis',
                      'Q_cars_p_CL_dis', 'Q_cars_p_CB_dis', 'Q_cars_p_DM_dis']
        # Loop through names of new outflow Timeseries objects
        for name in names_list:
            eles = name.split('_')
            # Find volume
            if eles[2] == 'p':
                volume = 500  # amount of water (L) pumped out per switch
            elif eles[2] == 'tb':
                # Get clean volume
                volume_c = self.get_timeseries({'dtype': 'V', 'dbase': 'man',
                                                'device': eles[2], 'layer': eles[3], 'state': 'c'})
                # Get dirty volume
                volume_d = self.get_timeseries({'dtype': 'V', 'dbase': 'man',
                                                'device': eles[2], 'layer': eles[3], 'state': 'd'})
                # combine both and take the mean
                volume = pd.concat([volume_c.data, volume_d.data]).values.mean()
            # Get corresponding switches Timeseries object
            switches = self.get_timeseries({'dtype': 'S', 'dbase': eles[1],
                                            'device': eles[2], 'layer': eles[3],
                                            'state': eles[4]})
            # Calculate discharge and add to database.
            discharge = switches.data * volume / area
            ts = Timeseries(name=name, data=discharge)
            self.add_timeseries(ts)

    def merge_timeseries(self, tsdic_main, tsdic_insert, params, name):
        '''Merge the data of two timeseries objects and add it to a new Timeseries object.'''
        # Get two Timeseries objects and make new one with the data of the main object
        ts_main = self.get_timeseries(tsdic_main)
        ts_ins = self.get_timeseries(tsdic_insert)
        ts_new = Timeseries(name, ts_main.data.copy())
        # Insert data from the other timeseries for given periods (params)
        if isinstance(params, list):
            for p in params:
                data_ins = ts_ins.get_data(period=p)
                ts_new.insert_values(data_ins)
        self.add_timeseries(ts_new)

    def calculate_Savg(self):
        '''Calculate average storage in cover soil.'''
        # Measurements used for topsoil and subsoil
        ts_TS = (self.get_timeseries({'dtype': 'sm', 'layer': '5cm'}) +
                 self.get_timeseries({'dtype': 'sm', 'layer': '15cm'}))
        ts_SS = (self.get_timeseries({'dtype': 'sm', 'layer': '40cm'}) + 
                 self.get_timeseries({'dtype': 'sm', 'layer': '80cm'}))
        depths = [200, 900]  # mm
        names = ['TS', 'SS']
        Savg = {}
        # For each layer calculate average water content and multiply with depth
        for ii, layer in enumerate([ts_TS, ts_SS]):
            th_layer = self.calculate_average_ts(layer)
            th_layer = th_layer / 100 * depths[ii]
            Savg[names[ii]] = th_layer
        Savg = pd.DataFrame(Savg)
        Savg = Savg.mean(axis=1)
        Savg = Timeseries('S_obs_mean', Savg)
        self.add_timeseries(Savg)

    def calculate_average_ts(self, ts_list):
        '''Calculate average data of data from inserted Timeseries object list.'''
        data = pd.DataFrame({ts.name: ts.get_data() for ts in ts_list})
        data = data.mean(axis=1)
        return data

    def correct_values(self, ts_dic, timerange=None, threshold_method=None, threshold_value=None,
                       correct_method=None, correct_value=None):
        """Correct data values using a method and value."""
        ts = self.get_timeseries(ts_dic)
        ts.correct_data(timerange, threshold_method, threshold_value, correct_method, correct_value)

    def define_outliers(self, ts_dic, params):
        """Define outliers for timeseries by giving a daterange or a threshold."""
        ts = self.get_timeseries(ts_dic)
        ts_list = ts if isinstance(ts, list) else [ts]
        for ts in ts_list:
            # If params is list of begin and end date of outliers, remove data between them
            if isinstance(params, list):
                ts.find_outliers_time(params)
            # Remove values which do not fulfill a condition made from method and threshold
            elif isinstance(params, dict):
                ts.find_outliers_condition(params)
            else:
                print(f"Parameter not valid: {params}.")

    def tb_weight_to_volume(self, rho_water=0.999):
        """Convert weight of water to volume."""
        # Loop through layers and clean and dirty volume
        for layer in ['CL', 'CB']:
            for state in ['c', 'd']:
                ts = self.get_timeseries({'dtype': 'V', 'dbase': 'man', 'layer': layer,
                                          'state': state})
                ts.data = ts.data / rho_water  # divide by density water.
