"""Docstring."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from classes.timeseries import Timeseries
from classes.error import ErrorMark

class Dataset():
    """Set of timeseries which have relation with each other."""

    # =============================================================================
    #     Constructor
    # =============================================================================

    def __init__(self, dbase):
        self.dbase = dbase  # Corresponding database
        self.timeseries = []  # List of Timeseries objects
        self.names_ts = []  # List of names of Timeseries objects

    def add_timeseries(self, ts_dic, period=None, resample=None, filter_gaps=None, cumulative=None):
        """Add all Timeseries objects that suffice search dictionary to this dataset."""
        # Get all Timeseries objects that suffice to search dictionary.
        # Else make list of single object.
        ts_list = self.dbase.get_timeseries(ts_dic)
        ts_list = ts_list if isinstance(ts_list, list) else [ts_list]
        for ts in ts_list:
            ts_ds = TimeseriesDataset(ts.name, ts.data, period, resample, filter_gaps, cumulative)
            self.timeseries.append(ts_ds)
            self.names_ts.append(ts_ds.name)

    def set_properties_timeseries(self, ts_dic=None, period=None, resample=None, filter_gaps=None, cumulative=None):
        """Set properties of all timeseries in this dataset complying with given dict."""
        # Get all ts complying with search criteria. If not given, edit all ts in this ds
        ts = self.get_timeseries(ts_dic) if ts_dic else self.timeseries
        ts_list = ts if isinstance(ts, list) else [ts]
        for ts in ts_list:
            ts.set_properties(period, resample, filter_gaps, cumulative)

    def get_timeseries(self, name_dic):
        """Return list of timeseries objects for given dtype, dbase, device, layer or state."""
        # Extract given dictionary: put given conditions in right order
        var = [[], [], [], [], [], []]
        var_str = ['name', 'dtype', 'dbase', 'device', 'layer', 'state']
        for i, v in enumerate(var_str):
            if v in name_dic.keys():
                var[var_str.index(v)] = name_dic[v]
        # List timeseries in dataset who comply with given search terms.
        ts_match = []
        for ts in self.timeseries:
            if ts.check_similarity(*var):
                ts_match.append(ts)
        # Return single object if only one timeseries matches, else return complete list.
        result = ts_match[0] if len(ts_match) == 1 else ts_match
        return result

    def get_dataframe(self, headers=None):
        '''Return dataframe containing data of all Timeseries objects in this dataset.'''
        ts_data = [ts.get_data_dataset() for ts in self.timeseries]
        data = dict(zip(self.names_ts, ts_data))
        data = pd.DataFrame(data=data)
        data.columns = headers if headers is not None else data.columns
        return data

    def get_intersection(self):
        """Return dataframe of intersect of given timeseries."""
        # Initiate lists
        data_list, index_list, name_list = [], [], []
        # Get data and index of each timeseries
        for i, ts in enumerate(self.timeseries):
            ts_data = ts.get_data_dataset()
            data_list.append(ts_data)
            index_list.append(ts_data.index)
            name_list.append(ts_data.name)
        # Get intersection of all indices
        index_int = index_list[0]
        for i in range(1, len(index_list)):
            index_int = np.intersect1d(index_int, index_list[i])
        # Get data of intersection and return
        data_int = np.array([data[index_int].values for data in data_list]).T
        df_int = pd.DataFrame(data=data_int, index=index_int, columns=name_list)
        return df_int

    def get_residual(self, period=None, fillna=False):
        '''Get residual of of difference of data in first column with sum of other columns.'''
        df = self.get_dataframe()
        # Select data from period
        if period:
            period_dt = pd.to_datetime(period)
            df = df[(df.index > period_dt[0]) & (df.index < period_dt[-1])]
        df = df.fillna(0) if fillna else df
        R = df[df.columns[0]]
        # Get difference of first column with sum of all other columns.
        for col in df.columns[1:]:
            R -= df[col]
        return R

    def get_residual_sum(self, period=None):
        '''Get difference of sum of first column with sum of other columns.'''
        df = self.get_dataframe()
        if period:
            period_dt = pd.to_datetime(period)
            df = df[(df.index > period_dt[0]) & (df.index < period_dt[-1])]
        R = np.sum(df[df.columns[0]].dropna().values)
        for col in df.columns[1:]:
            R -= np.sum(df[col].dropna().values)
        return R

    def group_timeseries_on_index(self, index):
        """Group ts_group on index of ts_index."""
        # Get index to group by
        if isinstance(index, dict):
            ts_index = self.get_timeseries(index)
            dates = ts_index.data.index  # Full index instead of subset to have no errors
            ts_list = self.timeseries.copy()
            ts_list.remove(ts_index)
        # External index
        else:
            dates = index
            ts_list = self.timeseries
        for ts in ts_list:
            if ts.resample is None:
                raise ErrorMark(f"Specify resample for timeseries {ts.name}.")
            ts.resample['index'] = dates
        df = self.get_dataframe()
        return df


class TimeseriesDataset(Timeseries):
    """Timeseries object usuable for Dataset."""

    def __init__(self, name, data, period, resample, filter_gaps, cumulative):
        super().__init__(name, data)
        self.period = period  # Store period to sample data on
        self.resample = resample  # Store resample parameters to sample data on
        self.filter_gaps = filter_gaps  # Store filtergap parameter to sample data on
        self.cumulative = cumulative  # Store if data needs to be returned cumulatively

    def set_properties(self, period, resample, filter_gaps, cumulative):
        self.period = period if period else self.period
        self.resample = resample if resample else self.resample
        self.filter_gaps = filter_gaps if filter_gaps else self.filter_gaps
        self.cumulative = cumulative if cumulative else self.cumulative

    def get_data_dataset(self):
        '''Return data according to inserted parameters for resampling.'''
        data = self.get_data(self.period, self.resample, self.filter_gaps, self.cumulative)
        return data
