"""Plot class. Contains objects that use matplotlib but are compatible with Timeseries and
Database objects."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from classes.error import ErrorMark
from classes.timeseries import Timeseries


class FigurePlot:
    """Plot class."""

    def __init__(self, dbase, rows=1, cols=1, **kwargs):
        self.dbase = dbase
        self.rows = rows
        self.cols = cols
        self.subplots = [Subplot(r, c) for r in range(rows) for c in range(cols)]
        self.cm = 1/2.54
        self.path = "C:\\Users\\vdbri\\Documents\\TUDELFT\\Afstuderen\\5_Paper\\figs"
        self.sizes = {'SMALL': 9,
                      'MEDIUM': 10,
                      'BIGGER': 12}
        self.kwargs = kwargs

    def plot_timeseries(self, ts_dic, subplot=None, axis=0, **kwargs):
        """Set properties of timeseries to plot."""
        ts = self.dbase.get_timeseries(ts_dic)
        if isinstance(ts, list):
            raise ErrorMark(f"Given ts_dic {ts_dic} not valid, returns {ts}.")
        sp_lst = self.__get_subplots(subplot, row=None, col=None)
        for sp in sp_lst:
            sp.add_timeseries_to_axis(axis, ts, kwargs)

    def plot_data(self, data=None, subplot=None, axis=0, **kwargs):
        sp_lst = self.__get_subplots(subplot)
        for sp in sp_lst:
            sp.add_data_to_axis(axis, data, kwargs)

    def __get_subplots(self, loc, row=None, col=None):
        """Get subplot with location."""
        if row is not None or col is not None:
            return [sp for sp in self.subplots if sp.check_loc(loc, row, col)]
        elif loc is None:
            return self.subplots
        elif isinstance(loc, list) and (self.rows == 1 or self.cols == 1):
            subplots = []
            for loc_i in loc:
                loc_ext = (0, loc_i) if self.rows == 1 else (loc_i, 0)
                subplots += [sp for sp in self.subplots if sp.check_loc(loc_ext, row, col)]
            return subplots
        elif isinstance(loc, list):
            subplots = []
            for loc_i in loc:
                subplots += [sp for sp in self.subplots if sp.check_loc(loc_i, row, col)]
            return subplots
        elif isinstance(loc, tuple):
            return [sp for sp in self.subplots if sp.check_loc(loc, row, col)]
        elif isinstance(loc, int) and (self.rows == 1 or self.cols == 1):
            loc_ext = (0, loc) if self.rows == 1 else (loc, 0)
            return [sp for sp in self.subplots if sp.check_loc(loc_ext, row, col)]
        else:
            raise ErrorMark(f"Location not valid: {loc}.")

    def set_layout_figure(self, **kwargs):
        self.kwargs = self.kwargs | kwargs

    def set_layout_subplot(self, row=None, col=None, subplot=None, **kwargs):
        sp_list = self.__get_subplots(subplot, row, col)
        for sp in sp_list:
            sp.kwargs = sp.kwargs | kwargs

    def set_layout_axis(self, subplot=None, row=None, col=None, axis=None, **kwargs):
        """Set layout of axis."""
        sp_list = self.__get_subplots(subplot, row, col)
        for sp in sp_list:
            sp.set_layout_axis(axis, kwargs)

    def set_properties_data(self, subplot=None, row=None, col=None, axis=None, series=None,
                            **kwargs):
        """Set layout of (group of) dataseries."""
        sp_list = self.__get_subplots(subplot, row, col)
        for sp in sp_list:
            sp.set_properties_data(axis, series, kwargs)

    def show(self):
        """Plot figure. Invokes similar method in subplots."""
        plt.rc('font', **{'size': self.sizes.get('SMALL'),
                          'family': 'sans-serif',
                          'sans-serif': 'Arial'})
        plt.rc('axes', titlesize=self.sizes.get('SMALL'),
                       labelsize=self.sizes.get('MEDIUM'))
        plt.rc('xtick', labelsize=self.sizes.get('SMALL'))
        plt.rc('ytick', labelsize=self.sizes.get('SMALL'))
        plt.rc('legend', fontsize=self.sizes.get('SMALL'))
        plt.rc('figure', titlesize=self.sizes.get('BIGGER'))
        plt.rc('lines', linewidth=1)
        figsize = tuple(ii * self.cm for ii in self.kwargs.get('figsize', (8, 6)))
        fig, axes = plt.subplots(nrows=self.rows, ncols=self.cols,
                                 sharex=self.kwargs.get('sharex', False),
                                 sharey=self.kwargs.get('sharey', False),
                                 figsize=figsize,
                                 squeeze=self.kwargs.get('squeeze', False),
                                 width_ratios=self.kwargs.get('width_ratios'),
                                 height_ratios = self.kwargs.get('height_ratios'),
                                 dpi=self.kwargs.get('dpi', 1000))
        self.figure, self.axes = fig, axes
        # Invoke show for subplots; Make legend for whole figure if relevant
        if 'legend_loc' in self.kwargs:
            lines, labels = ([], [])
            for sp in self.subplots:
                legend_handles = sp.show_sp(axes, legend=False)
                if legend_handles[0]:
                    lines += legend_handles[0]
                    labels += legend_handles[1]
            fig.legend(lines, labels, bbox_to_anchor=self.kwargs['legend_loc'])
        else:
            for sp in self.subplots:
                sp.show_sp(axes, legend=True)
        # Plot layout and return plotted data
        self.__plot_layout(fig, axes)
        return [sp.get_data() for sp in self.subplots]

    def __plot_layout(self, fig, axes):
        # Place text in figure
        if 'text' in self.kwargs.keys():
            if isinstance(self.text, list):
                for ii, t in enumerate(self.text):
                    fig.text(self.text_loc[ii][0], self.text_loc[ii][1], t)
            else:
                fig.text(self.text_loc[0], self.text_loc[1], self.text)
        # Title
        fig.suptitle(self.kwargs.get('title'))
        # Spacing
        if ('hspace' in self.kwargs) or ('wspace' in self.kwargs):
            fig.subplots_adjust(hspace=self.kwargs.get('hspace'),
                                wspace=self.kwargs.get('wspace'))
        else:
            fig.tight_layout()
        fig.align_labels()


class Subplot:
    """Subplot class."""

    def __init__(self, row, col):
        self.row = row
        self.col = col
        self.n_axes = 1
        self.axes = [Axis()]
        self.kwargs = {}

    def check_loc(self, loc, row, col):
        """Return if this subplot is on the given location."""
        if row is not None:
            return row == self.row
        elif col is not None:
            return col == self.col
        else:
            return loc[0] == self.row and loc[1] == self.col

    def add_axis(self, axis):
        """Add extra y-axis to this subplot."""
        self.n_axes = axis + 1
        while len(self.axes) < self.n_axes:
            axis = Axis()
            self.axes.append(axis)

    def add_timeseries_to_axis(self, axis, timeseries, kwargs):
        """Add timeseries to axis."""
        if axis > self.n_axes - 1:
            self.add_axis(axis)
        ts_plot = DataseriesPlot(timeseries, kwargs)
        self.axes[axis].add_series(ts_plot)

    def add_data_to_axis(self, axis, data, kwargs):
        if axis > self.n_axes - 1:
            self.add_axis(axis)
        ds_plot = DataseriesPlot(data, kwargs)
        self.axes[axis].add_series(ds_plot)

    def get_data(self):
        """Return list of data in this subplot."""
        data = []
        for ax in self.axes:
            data += ax.get_data()
        return data

    def set_layout_axis(self, axis, kwargs):
        """Set layout of axis."""
        axes = self.__get_axes(axis)
        for ax in axes:
            ax.set_layout(kwargs)

    def set_properties_data(self, axis, series, kwargs):
        """Set properties of data in axis."""
        axes = self.__get_axes(axis)
        for ax in axes:
            ax.set_properties_data(series, kwargs)

    def __get_axes(self, axis):
        if axis is None:
            return self.axes
        elif isinstance(axis, list):
            return axis
        elif isinstance(axis, int):
            return [self.axes[axis]]

    def show_sp(self, axes, legend):
        """Plot this subplot by invoking show methods of axis. Also make legend."""
        axis = [axes[self.row, self.col]]
        # Make new axis if self.n_axes is larger than 1
        if self.n_axes > 1:
            axis = axis + [axis[0].twinx() for i in range(self.n_axes - 1)]
        # Plot axes and get handles to make legend for whole subplot
        lines, labels = [], []
        for i, ax in enumerate(axis):
            legend_handles = self.axes[i].show_ax(ax)
            if legend_handles[0]:
                lines += legend_handles[0]
                labels += legend_handles[1]
        if legend and lines and legend and self.kwargs.get('legend_loc') != 'None':
            ax.legend(lines, labels,
                      loc=self.kwargs.get('legend_loc', 'best'),
                      ncol=self.kwargs.get('legend_ncol', 1))
        elif not legend:
            return lines, labels  # Return lines and labels for figure legend


class Axis():
    """Axis class."""

    def __init__(self):
        self.series = []
        self.bar = {}
        self.kwargs = {}

    def add_series(self, series):
        """Add dataseries to this axis."""
        self.series.append(series)
        if 'bar' in series.kwargs:
            self.bar[series] = series.kwargs.get('bar_stack', 0)

    def get_data(self):
        """Get data in this axis."""
        return [sr.get_data_plot() for sr in self.series]

    def set_layout(self, kwargs):
        """Set layout variables of axis."""
        self.kwargs = self.kwargs | kwargs

    def set_properties_data(self, series, kwargs):
        """Set properties of dataseries in this axis, all, subset or single."""
        series_lst = self.__get_series(series)
        for i, sr in enumerate(series_lst):
            sr.set_properties(kwargs)
            if 'bar' in sr.kwargs:
                self.bar[sr] = sr.kwargs.get('bar_stack', 0)

    def __get_series(self, series):
        if series is None:
            return self.series
        elif isinstance(series, list):
            return [self.series[ii] for ii in series]
        elif isinstance(series, int):
            return [self.series[series]]

    def show_ax(self, ax):
        """Plot this axis. Return legend handles to subplot."""
        self.__plot_data(ax)
        self.__plot_layout(ax)
        return ax.get_legend_handles_labels()

    def __plot_data(self, ax):
        if 'boxplot' in self.kwargs:
            # Get data
            data = [sr.get_data_plot().values for sr in self.series]
            labels = [sr.kwargs.get('label') for sr in self.series]
            ax.boxplot(data, labels=labels, widths=self.kwargs.get('boxplot_width', 0.5))
        elif self.bar:
            stacks = set(self.bar.values())
            width_bar = self.kwargs.get('bar_width', 0.6) / float(len(stacks))
            bottom = dict(zip(stacks, [0] * len(stacks)))
            for sr, stack in self.bar.items():
                if self.kwargs.get('bar_align') == 'left':
                    dx = stack * width_bar
                elif self.kwargs.get('bar_align') == 'right':
                    dx = -self.kwargs.get('bar_width', 0.6) + stack * width_bar
                else:
                    dx = -self.kwargs.get('bar_width', 0.6) / 2 + stack * width_bar
                bottom[stack] += sr.show(ax, width_bar, dx, bottom[stack])
            for sr in self.series:
                if sr not in self.bar.keys():
                    d = sr.show(ax)
        else:
            for sr in self.series:
                d = sr.show(ax)
        if 'plot_mean' in self.kwargs:
            self.__plot_mean(ax)


    def __plot_mean(self, ax):
        """Plot mean of data (list, int, or all) in this axis."""
        plot_mean = self.kwargs.get('plot_mean')
        # Plot mean of subset of dataseries
        if isinstance(plot_mean, list):
            mean = np.concatenate([self.series[i].get_data_plot().values for i in plot_mean]).mean()
        elif isinstance(plot_mean, int):
            mean = self.series[plot_mean].get_data_plot().values.mean()
        # Plot mean of all dataseries
        else:
            mean = np.concatenate([sr.get_data_plot().values for sr in self.series]).mean()
        # ax.axhline(mean, color='k', label='Mean')
        ax.axhline(mean, color='k', label='Gemiddelde')

    def __plot_layout(self, ax):
        ax.set_ylabel(self.kwargs.get('ylabel'),
                      loc=self.kwargs.get('ylabel_loc', 'center'))
        ax.set_ylim(self.kwargs.get('ylim'))
        ax.set_yscale(self.kwargs.get('yscale', 'linear'))
        if 'yticks' in self.kwargs:
            ax.set_yticks(self.kwargs.get('yticks'))
        ax.set_xlabel(self.kwargs.get('xlabel'),
                      loc=self.kwargs.get('xlabel_loc', 'center'))
        ax.set_xlim(self.kwargs.get('xlim'))
        self.__set_xticks(ax)
        ax.set_aspect(self.kwargs.get('aspect', 'auto'))
        ax.set_title(self.kwargs.get('title'), fontsize=13, loc=self.kwargs.get('title_loc'))
        if isinstance(self.kwargs.get('grid_linestyle'), dict):
            ax.grid(which='minor', linestyle=self.kwargs.get('grid_linestyle')['minor'],
                    color=self.kwargs.get('grid_color', 'silver'),
                    zorder=0)
            ax.grid(which='major', linestyle=self.kwargs.get('grid_linestyle')['major'],
                    color=self.kwargs.get('grid_color', 'silver'),
                    zorder=0)
        else:
            ax.grid(linestyle=self.kwargs.get('grid_linestyle', '-'),
                    color=self.kwargs.get('grid_color', 'silver'),
                    zorder=0)
        if self.kwargs.get('invert_yaxis'):
            ax.invert_yaxis()
        if isinstance(self.kwargs.get('text'), list):
            for ii, txt in enumerate(self.kwargs.get('text')):
                ax.text(self.kwargs.get('text_loc')[ii][0],
                        self.kwargs.get('text_loc')[ii][1], txt,
                        transform=ax.transAxes, bbox={'facecolor': 'white', 'alpha': 0.5})
        elif self.kwargs.get('text') is not None:
            ax.text(self.kwargs.get('text_loc')[0], self.kwargs.get('text_loc')[1],
                    self.kwargs.get('text'),
                    transform=ax.transAxes, bbox={'facecolor': 'white', 'alpha': 0.5})

    def __set_xticks(self, ax):
        xticks = self.kwargs.get('xticks')
        if isinstance(xticks, dict):
            if 'ticks' in xticks and 'labels' in xticks:
                ax.set_xticks(ticks=xticks.get('ticks'), labels=xticks.get('labels'),
                              rotation=xticks.get('rotation', 0))
            else:
                ax.xaxis.set_major_locator(xticks.get('loc_maj', ax.xaxis.get_major_locator()))
                ax.xaxis.set_major_formatter(xticks.get('for_maj', ax.xaxis.get_major_formatter()))
                ax.xaxis.set_minor_locator(xticks.get('loc_min', ax.xaxis.get_minor_locator()))
                ax.xaxis.set_minor_formatter(xticks.get('for_min', ax.xaxis.get_minor_formatter()))
                ax.tick_params(axis='x', which='major', pad=xticks.get('pad_maj', 3.5),
                               labelrotation=self.kwargs.get('xlabel_rot', 0))
        elif xticks is not None:
            ax.set_xticks(ticks=xticks,
                           labels=xticks,
                          # labels=[f"%.1f" % label for label in xticks],
                          rotation=self.kwargs.get('xlabel_rot', 0),
                          ha=self.kwargs.get('xlabel_ha', 'center'))
        elif self.kwargs.get('xlabel_rot') is not None:
            ax.set_xticks(ticks=ax.get_xticks(),
                          labels=ax.get_xticklabels(),
                          rotation=self.kwargs.get('xlabel_rot'), #rotation_mode='anchor',
                          ha=self.kwargs.get('xlabel_ha', 'center'))


class DataseriesPlot():
    """Plotting dataseries."""

    def __init__(self, data, kwargs):
        self.data = data
        self.kwargs = kwargs

    def get_data_plot(self):
        if isinstance(self.data, Timeseries):
            return self.data.get_data(self.kwargs.get('period'),
                                      self.kwargs.get('resample'),
                                      self.kwargs.get('filter_gaps'),
                                      self.kwargs.get('cumulative'))
        else:
            return self.data
        

    def set_properties(self, kwargs):
        self.kwargs = self.kwargs | kwargs

    def show(self, ax, width_bar=None, dx=None, bottom=None):
        data = self.get_data_plot()
        # Plot as bar
        if 'bar' in self.kwargs:
            # If previous data was plotted negatively and current data is positive,
            # reset bottom to 0
            if isinstance(bottom, np.ndarray):
                bottom[(bottom < 0) & (data.values > 0)] = 0
            edgecolor = 'None' if self.kwargs.get('hatch') is None else 'k'
            ax.bar(x=self.__get_xdata(data, dx),
                   height=data.values, bottom=bottom,
                   align='edge', width=width_bar, ec=edgecolor,
                   label=self.kwargs.get('label'),
                   color=self.kwargs.get('color'),
                   hatch=self.kwargs.get('hatch'))
        # Plot with fill between index and values
        elif 'fill_data' in self.kwargs:
            ax.fill_between(data.index, data.values,
                            self.kwargs.get('fill_data'),
                            self.kwargs.get('fill_color'),
                            self.kwargs.get('fill_alpha'))
        # Plot as line using style
        elif 'style' in self.kwargs:
            ax.plot(data, self.kwargs.get('style'), label=self.kwargs.get('label'))
        # Plot as line using specific arguments
        else:
            ax.plot(data,
                    label=self.kwargs.get('label'),
                    color=self.kwargs.get('color'),
                    marker=self.kwargs.get('marker'),
                    markersize=self.kwargs.get('markersize'),
                    linestyle=self.kwargs.get('linestyle'),
                    linewidth=self.kwargs.get('linewidth'))
        # Plot measurements with NaN values as grey bars
        if self.kwargs.get('plot_nan'):
            # Get data points with NaN values
            nodata = data.copy()
            nodata = nodata[nodata.isna()]
            # Plot NaN values as grey bars extending from the negative maximum of 
            # the data to 10 times this maximum. Correct with specifying ylimits.
            if len(nodata) > 0:
                nodata = nodata.fillna(data.dropna().values.max()*10)
                ax.bar(x=self.__get_xdata(nodata, dx),
                       height=2*nodata.values, bottom=-nodata.values,
                       align='edge', width=width_bar,
                       ec='None', color='lightgrey', alpha=0.5, label='No data')
        return data.values

    def __get_xdata(self, data, dx):
        if dx is not None and isinstance(data.index[0], pd.Timestamp):
            return data.index + pd.Timedelta(dx, unit='d')
        elif dx is not None:
            return data.index + dx
        else:
            return data.index
