"""Docstring."""

import numpy as np
import pandas as pd

from classes.error import ErrorMark
import functions.hydroyear as hy


# %% Timeseries class

class Timeseries:
    """Superclass timeseries."""

    def __init__(self, name, data):
        self.name = name
        self.__set_attributes_from_name()
        self.data = data

    def __set_attributes_from_name(self):
        """Set the attributes derived from the name of the object."""
        attributes = self.name.split('_')
        attributes += [''] * (5 - len(attributes))
        self.dtype = attributes[0]
        self.dbase = attributes[1]
        self.device = attributes[2]
        self.layer = attributes[3]
        self.state = attributes[4]

    def check_similarity(self, name, dtype, dbase, device, layer, state):
        """Check if this timeseries object complies with gives search conditions."""
        l_cond = [name, dtype, dbase, device, layer, state]  # list of properties to check on
        l_cond_str = ['name', 'dtype', 'dbase', 'device', 'layer', 'state']
        # Check how much conditions are given
        beg = sum([len(cond) != 0 for cond in l_cond])  # Number of conditions to meet
        # Check for each given condition if it is the same as this timeseries object (in name)
        fin = []  # List containing conditions that are met
        for i, cond in enumerate(l_cond):
            # If single condition for property is given
            if isinstance(cond, str):
                fin.append(eval("cond == self." + l_cond_str[i]))
            # If multiple conditions for properties are given
            elif isinstance(cond, list):
                fin.append(eval("self." + l_cond_str[i] + " in cond"))
            else:
                raise ErrorMark(f"Input not valid: {cond}")
        # Check if the amount of true checks is the same as the amount of conditions given
        ch_tot = sum(fin) == beg if beg != 0 else None
        return ch_tot

    def get_data(self, period=None, resample=None, filter_gaps=None, cumulative=False):
        """Return data of given year, month, day, resampled to the desired scale."""
        data, begin_end = self.__filter_period(period)  # Filter data on given period
        data = self.__resample_data(data, resample, begin_end, filter_gaps)  # Resample and filter data
        data = data.cumsum() - data.dropna()[0] if cumulative else data  # Cumsum, setting first value to 0
        return data

    def __filter_period(self, period):
        """Return masked data for a year, month or day or between two dates.
        If given a day, month and year should also be given.
        """
        index = self.data.index
        if period is not None:
            # Convert dictionary with period to list
            if isinstance(period, dict):
                begin_end = self.__get_period_dic(period)
            # Period is already of pd Timestamp type
            elif isinstance(period[0], pd.Timestamp):
                begin_end = period
            # Convert to pd Timestamp type
            else:
                begin_end = pd.to_datetime(period)
            # Mask data with given period
            mask = index[(index >= begin_end[0]) & (index <= begin_end[1])]
            data = self.data.loc[mask]
            # Return data and given begin and and
            return [data, begin_end]
        else:
            # Return data without filtering
            begin_end = [index[0], index[-1]]
            return [self.data, begin_end]

    def __get_period_dic(self, period):
        """Return a list containing two datetime objects of the begin and end of given period."""
        year = period['year'] if 'year' in period.keys() else None
        month = period['month'] if 'month' in period.keys() else None
        day = period['day'] if 'day' in period.keys() else None
        if day:
            begin = pd.to_datetime(str(year) + '-' + str(month) + '-' + str(day))
            end = pd.to_datetime(str(year) + '-' + str(month) + '-' + str(day + 1)) - pd.Timedelta(1)
        elif month:
            begin = pd.to_datetime(str(year) + '-' + str(month))
            end = pd.to_datetime(str(year) + '-' + str(month + 1)) - pd.Timedelta(1)
        else:
            begin = pd.to_datetime(str(year))
            end = pd.to_datetime(str(year + 1)) - pd.Timedelta(1)
        return [begin, end]

    def __resample_data(self, data, resample, begin_end, filter_gaps):
        """Return resampled data."""
        if resample:
            # Unpack dictionary
            scale = resample['scale'] if 'scale' in resample.keys() else None
            index = resample['index'] if 'index' in resample.keys() else None
            method = resample['method'] if 'method' in resample.keys() else None
            # Group data and sum the bins
            data_grouped, data_sum = self.__group_data(data, index, scale, begin_end)
            # Filter out invalid data groups
            if filter_gaps:
                data_sum = self.__filter_gaps(data_sum, data_grouped, filter_gaps,
                                              scale, begin_end, index)
            # Calculate group aggregates
            data = self.__calculate_group_aggregate(data_sum, data_grouped, filter_gaps, method, scale)
            # Fill begin and end with NaN values if given begin and end extend beyond
            # data range
            if index is None:
                data = self.__fill_begin_end(data, begin_end, scale)
        return data

    def __group_data(self, data, index, scale, begin_end):
        """Group given data on scale or index."""
        # Group data between dates in given index
        if index is not None:
            data_grouped = data.groupby(pd.cut(data.index, index))  # Cut data on index values
            data_sum = data_grouped.sum()
            data_sum.index = pd.IntervalIndex(data_sum.index).right  # Set date as end of interval
            data_sum.index.name = 'datetime'
        elif scale:
            data_grouped = data.groupby(pd.Grouper(freq=scale))  # Group data on scale
            data_sum = data_grouped.sum()
        else:
            raise ErrorMark(f"Input resampling not valid: {index}")
        return (data_grouped, data_sum)

    def __filter_gaps(self, data_sum, data_grouped, filter_gaps, scale, begin_end, index):
        """Filter out data groups if they or their predecessor contain a datagap larger than given."""
        # Get begin and end of period and maximum gap
        gap_max = self.__get_gap_max(filter_gaps, scale)
        # Get first and last measurement and total count of each data group
        group_begin = data_grouped.head(1).index
        group_end = data_grouped.tail(1).index
        group_count = data_grouped.count()
        # Loop through all datagroups
        del_next = False
        j = 0  # Counter for group_begin and group_end, they only contain values of non-zero groups
        for i in range(len(data_sum)):
            # Delete group if it does not contain measurements
            if group_count[i] == 0:
                data_sum.iloc[i] = np.nan
                del_next = True
                j -= 1
            # Delete if previous group had a gap to prevent an accumulation peak.
            elif del_next:
                data_sum.iloc[i] = np.nan
                del_next = False
            # Delete first measurement if period between start and first point is too large
            elif (i == 0) and (group_begin[j] - begin_end[0] > gap_max):
                data_sum.iloc[i] = np.nan
            else:
                end_period = self.__get_end_period( i, data_grouped, data_sum,
                                                   scale, index, begin_end)
                # Delete if time difference between end of period and last measurement is too large
                if end_period - group_end[j] > gap_max:
                    data_sum.iloc[i] = np.nan
                    del_next = True
            j += 1
        return data_sum

    def __get_gap_max(self, filter_gaps, scale):
        """Return Timedelta object of maximal difference in time."""
        gap_max_rel = filter_gaps['gap_max_rel'] if 'gap_max_rel' in filter_gaps.keys() else None
        if scale[-1] == 'M':
            n, unit = (30*int(scale[:-1]), 'D')  # use days as time scale
        elif scale[-1] == 'Y':
            n, unit = (365, 'D')  # use days as time scale
        else:
            n, unit = (int(scale[:-1]), scale[-1])
        gap_max = pd.Timedelta(value=n*gap_max_rel, unit=unit)  # Return as pd Timedelta object
        return gap_max
    
    def __get_end_period(self, i, data_grouped, data_sum, scale, index, begin_end):
        # If this is the last of the groups, end of period is defined end.
        if i == len(data_grouped) - 1:
            return begin_end[1]
        # If resampled on index, index of bin is end of period
        elif index is not None:
            return data_sum.index[i]
        # If yearly scale, next year is end of period
        elif scale[-1] == 'Y':
            return pd.to_datetime(str(data_sum.index[i+1].year))
        # If monthly scale, next month is end of period
        elif scale[-1] == 'M':
            return pd.to_datetime(str(data_sum.index[i+1].year) +
                                        '-' + str(data_sum.index[i+1].month))
        else:
            return data_sum.index[i+1]

    def __calculate_group_aggregate(self, data_sum, data_grouped, filter_gaps, method, scale):
        if not filter_gaps:
            data_sum = data_sum[data_grouped.count() != 0]
        if method == 'avg':
            count = data_grouped.count()
            count = self.__reset_intervalindex(count) if isinstance(count.index[0], pd.Interval) else count
            data_agg = (data_sum / count)
            data_agg.name = self.name
        elif method == 'sum':
            data_agg = data_sum
            data_agg.name = self.name
        elif method == 'first':
            data_agg = data_grouped.first()
            data_agg = self.__reset_intervalindex(data_agg) if isinstance(data_agg.index[0], pd.Interval) else data_agg
        elif method == 'last':
            data_agg = data_grouped.last()
            data_agg = self.__reset_intervalindex(data_agg) if isinstance(data_agg.index[0], pd.Interval) else data_agg
        elif method == 'change':
            first = data_grouped.first()
            last = data_grouped.last()
            if isinstance(first.index[0], pd.Interval):
                first = self.__reset_intervalindex(first)
                last = self.__reset_intervalindex(last)
            data_agg = last - first
        elif method == 'max':
            data_agg = data_grouped.max()
            data_agg = self.__reset_intervalindex(data_agg) if isinstance(data_agg.index[0], pd.Interval) else data_agg
        elif method == 'max_rel':
            data_agg = data_grouped.max() / data_grouped.sum()
            data_agg = self.__reset_intervalindex(data_agg) if isinstance(data_agg.index[0], pd.Interval) else data_agg
        elif method == 'std':
            data_agg = data_grouped.std()
            data_agg = self.__reset_intervalindex(data_agg) if isinstance(data_agg.index[0], pd.Interval) else data_agg
        else:
            raise ErrorMark(f'Resamping method not valid: {method}.')
        if scale[-1] == 'Y':
            data_agg.index = hy.get_hydroyear_from_index(data_agg.index)
        return data_agg

    def __reset_intervalindex(self, series):
        series.index = pd.IntervalIndex(series.index).right
        series.index.name = 'datetime'
        return series

    def __fill_begin_end(self, data, begin_end, scale):
        '''Fill values between given begin and end with NaN values.'''
        drange = pd.date_range(start=begin_end[0].strftime('%Y-%m-%d'),
                               end=begin_end[1].strftime('%Y-%m-%d'),
                               freq=scale)
        dates_nan = drange.difference(data.index)
        data_nan = pd.Series(index=dates_nan, data=np.full(shape=len(dates_nan),
                                                           fill_value=np.nan))
        data = pd.concat([data, data_nan]).sort_index()
        return data
    

    # %% Data cleaning

    def cumulative_to_discrete_data(self):
        """Change cumulative data values to discrete values while taking into account resets of the cumulative data."""
        # Get data
        index_cum = self.data.index
        data_cum = self.data.values
        # Delete readings that have another reading within 1 minute
        delta_t = index_cum[1:] - index_cum[:-1]
        mask = (delta_t < pd.Timedelta(1, unit='m'))
        index_cum = np.array([*index_cum[:-1][~mask], index_cum[-1]])
        data_cum = np.array([*data_cum[:-1][~mask], data_cum[-1]])
        # Correct cumulative count resets
        delta_v = data_cum[1:] - data_cum[:-1]
        mask_neg = delta_v < 0
        mask_max = data_cum[:-1] > 50000
        mask_auto = mask_neg * mask_max  # automatic resets because max int is reached
        mask_man = mask_neg * ~mask_max  # manual resets before max int value is reached
        delta_v[mask_auto] = 65535 - data_cum[:-1][mask_auto] + data_cum[1:][mask_auto]
        delta_v[mask_man] = data_cum[1:][mask_man]
        # Discrete data
        data_dis = np.array([0, *delta_v])
        self.data = pd.Series(data=data_dis, index=index_cum)

    def correct_data(self, timerange, threshold_method, threshold_value, correct_method, correct_value):
        # Find data
        if timerange:
            timerange = pd.to_datetime(timerange)
            index_wrong = self.data.index[(self.data.index > timerange[0]) & (self.data.index < timerange[-1])]
        elif threshold_method and threshold_value:
            index_wrong = self.data.index[eval('self.data' + threshold_method + 'threshold_value')]
        else:
            raise ErrorMark("Method of data correction not given right.")
        # Correct data
        if correct_method:
            self.data[index_wrong] = eval('self.data[index_wrong]' + correct_method + 'correct_value')
        else:
            self.data[index_wrong] = correct_value

    def find_outliers_time(self, time_list):
        if len(time_list) == 2:
            begin, end = [pd.to_datetime(par) for par in time_list]
            self.data[(self.data.index > begin) & (self.data.index < end)] = np.nan
        else:
            time_arr = np.array(time_list)
            self.data[self.data.index.intersection(time_arr)] = np.nan
        self.data = self.data[self.data.notna()]

    def find_outliers_condition(self, condition):
        """Find and set outliers based on a condition (logical operation)."""
        threshold = condition['threshold'] if 'threshold' in condition.keys() else None
        method = condition['method'] if 'method' in condition.keys() else None
        self.data[eval('self.data' + method + 'threshold')] = np.nan
        self.data = self.data[self.data.notna()]

    def insert_values(self, data_ins):
        self.data = pd.concat([self.data, data_ins])
        self.data = self.data.sort_index()
