"""File with functions used for data analysis."""

import os
import numpy as np
import pandas as pd
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.cm import ScalarMappable

import functions.hydroyear as hy
from classes.database import Database
from classes.dataset import Dataset
from classes.timeseries import Timeseries
from classes.plot import FigurePlot


def load_database(path):
    db = Database()  # initialise database object
    fnames = os.listdir(path)  # Get names of files at location
    for file in fnames:
        # Only read the csv files
        if file[-3:] == 'csv':
            data = pd.read_csv(path+file, index_col=0, parse_dates=True)
            # Convert each column in the dataframe to a Timeseries object
            # and add to database
            for col in data.columns:
                ts = Timeseries(name=col, data=data[col].dropna())
                db.add_timeseries(ts)
    return db


def plot_climate(db):
    # Get data
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'P', 'dbase': 'knmiDK'})
    ds.add_timeseries({'dtype': 'ETref', 'dbase': 'knmiDK'})
    ds.add_timeseries({'dtype': 'Tatm', 'dbase': 'knmiDK'})
    ds.set_properties_timeseries(period=['1980-04', '2010-03-31'])
    data = ds.get_dataframe()
    # Average by month
    P, ETref, Tatm = data.columns
    datap = data.groupby(pd.Grouper(freq='M')).sum()
    datap = datap.groupby(by=datap.index.month).mean()
    datap[Tatm] = data[Tatm].groupby(by=data.index.month).mean()
    months = pd.to_datetime(datap.index, format='%m').strftime('%b')
    # Plot
    fig = FigurePlot(db, figsize=(14, 8))
    fig.plot_data(datap[P],
                  bar=True, bar_stack=0,
                  color='C0', hatch='...',
                  label='Precipitation')
    fig.plot_data(datap[ETref],
                  color='C1', hatch='xxx',
                  bar=True, bar_stack=1,
                  label='ET$_\mathrm{ref}$')
    fig.plot_data(axis=1, data=datap[Tatm],
                  color='C3', marker='s', linewidth=2,
                  label='Temperature')
    fig.set_layout_axis(axis=0,
                        yticks=np.arange(0, 120, 10), ylim=(0, 110),
                        bar_width=0.8,
                        ylabel='Average precipitation and\nET$_\mathrm{ref}$ (mm/month)',
                        xticks={'ticks': datap.index,
                                'labels': months,})
                                # 'rotation': 90})
    fig.set_layout_axis(axis=1, yticks=np.arange(0, 24, 2),
                        grid_linestyle='None',
                        ylabel='Average monthly temperature (°C)')
    fig.set_layout_subplot(legend_loc='upper left')
    d = fig.show()
    return d


def plot_climate_studyperiod(db):
    # Get data
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'P', 'dbase': 'knmiBerkhout'})
    ds.add_timeseries({'dtype': 'ETref', 'dbase': 'knmiBerkhout'})
    ds.add_timeseries({'dtype': 'Tatm', 'dbase': 'knmiBerkhout'})
    ds.set_properties_timeseries(period=['2010-04', '2023-03-31'])
    data = ds.get_dataframe()
    P, ETref, Tatm = data.columns
    # Average by month
    datap = data.groupby(pd.Grouper(freq='M')).sum()
    datap = datap.groupby(by=datap.index.month).mean()
    datap[Tatm] = data[Tatm].groupby(by=data.index.month).mean()
    months = pd.to_datetime(datap.index, format='%m').strftime('%b')
    # Plot
    fig = FigurePlot(db, figsize=(14, 8))
    fig.plot_data(datap[P],
                  bar=True, bar_stack=0,
                  color='C0', hatch='...',
                  label='Precipitation')
    fig.plot_data(datap[ETref],
                  color='C1', hatch='xxx',
                  bar=True, bar_stack=1,
                  label='ET$_\mathrm{ref}$')
    fig.plot_data(axis=1, data=datap[Tatm],
                  color='C3', marker='s', linewidth=2,
                  label='Temperature')
    fig.set_layout_axis(axis=0,
                        yticks=np.arange(0, 120, 10), ylim=(0, 110),
                        bar_width=0.8,
                        ylabel='Average precipitation and\nET$_\mathrm{ref}$ (mm/month)',
                        xticks={'ticks': datap.index,
                                'labels': months,})
                                # 'rotation': 90})
    fig.set_layout_axis(axis=1, yticks=np.arange(0, 24, 2),
                        grid_linestyle='None',
                        ylabel='Average monthly temperature (°C)')
    fig.set_layout_subplot(legend_loc='upper left')
    d = fig.show()
    return d


def plot_Q_annual(db, period=['2010-04-01', '2023-03-31']):
    index = hy.get_index_hydroyear()
    # Plot
    fig = FigurePlot(db, cols=2, figsize=(14, 8),
                     legend_loc=(0.5, 0.2), width_ratios=[2, 1])
    # Plot data
    fig.plot_timeseries({'dtype': 'P', 'dbase': 'knmiBerkhout'},
                        subplot=1,
                        resample={'method': 'sum', 'scale': '1Y', 'index': index},
                        label='Precipitation')
    fig.plot_timeseries(ts_dic={'dtype': 'Q', 'device': 'com', 'layer': 'CB'},
                        resample={'method': 'sum', 'scale': '1Y', 'index': index},
                        bar_stack=0, bar=True,
                        color='C3', label='Breakthrough', hatch='----')
    fig.plot_timeseries(ts_dic={'dtype': 'Q', 'device': 'com', 'layer': 'CL'},
                        resample={'method': 'sum', 'scale': '1Y', 'index': index},
                        bar_stack=0, bar=True,
                        color='C2', label='Diverted infiltration', hatch='////')
    fig.set_properties_data(period=period)
    # Plot residual
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'P', 'dbase': 'knmiBerkhout'},
                      resample={'method': 'sum', 'scale': '1Y', 'index': index})
    for ts in ['CB', 'CL']:
        ds.add_timeseries(ts_dic={'dtype': 'Q', 'device': 'com', 'layer': ts},
                          resample={'method': 'sum', 'scale': '1Y', 'index': index})
    ds.set_properties_timeseries(period=period)
    res = ds.get_residual()
    fig.plot_data(data=res,
                  color='C9', label='Evapotranspiration', hatch='\\\\\\\\',
                  bar_stack=0, bar=True)
    # Layout
    fig.set_layout_axis(subplot=0, bar_width=0.8, yticks=np.arange(0, 1200, 100),
                        xticks=np.arange(2010, 2023), xlabel_rot=90,
                        ylabel='Flux (mm/a)',
                        text='Hydrological year',
                        text_loc=(0.32, -0.4))
    fig.set_layout_axis(subplot=1, boxplot=True, xlabel_rot=90, xlabel_ha='right',
                        yticks=np.arange(0, 1200, 100), ylim=(0, 1100),
                        boxplot_width=0.7)
    data = fig.show()
    return data



def plot_Q_daily(db, period=['2009-10', '2023-04']):
    fig = FigurePlot(db, rows=4, figsize=(19, 16), sharex=True,
                     height_ratios=[4, 3, 3, 3])
    # Plot data
    fig.plot_timeseries({'dtype': 'Q', 'device': 'tb', 'layer': 'CL'},
                        subplot=2, color='C2')
    fig.plot_timeseries({'dtype': 'Q', 'device': 'tb', 'layer': 'CB'},
                        subplot=3, color='C3')
    fig.set_properties_data(resample={'method': 'sum', 'scale': '1D'},
                            filter_gaps={'gap_max_rel': 0.08})
    fig.plot_timeseries({'dtype': 'P', 'dbase': 'knmiBerkhout'},
                        subplot=0, axis=1, color='C0',
                        resample={'method': 'sum', 'scale': '1D'})
    fig.plot_timeseries({'dtype': 'ETref', 'dbase': 'knmiBerkhout'},
                        subplot=0, axis=0, color='C1',
                        resample={'method': 'sum', 'scale': '1D'})
    fig.set_properties_data(bar=True, plot_nan=True)
    fig.plot_timeseries({'dtype': 'sm', 'device': 'up', 'layer': '80cm'},
                        subplot=1,
                        color='C5', label='Upslope', linestyle='-',
                        resample={'method': 'avg', 'scale': '1D'})
    fig.plot_timeseries({'dtype': 'sm', 'device': 'mid', 'layer': '80cm'},
                        subplot=1,
                        color='C7', label='Midslope', linestyle='--',
                        resample={'method': 'avg', 'scale': '1D'})
    fig.plot_timeseries({'dtype': 'sm', 'device': 'down', 'layer': '80cm'},
                        subplot=1,
                        color='C9', label='Downslope', linestyle=':',
                        resample={'method': 'avg', 'scale': '1D'})
    fig.set_properties_data(period=period, linewidth=1)
    fig.plot_data(subplot=1, 
                  data=pd.Series(index=pd.to_datetime(2*['2013-09']),
                                 data=[0, 30]),
                  linestyle='--', color='k')
    # Layout
    fig.set_layout_axis(bar_width=1,
                        bar_align='right',
                        grid_linestyle='--')
    fig.set_layout_axis(subplot=0, axis=0,
                        ylim=(0, 15), yticks=np.arange(0, 6, 1),
                        ylabel='$\mathregular{ET_{ref}}$ (mm/d)', ylabel_loc='bottom')
    fig.set_layout_axis(subplot=0, axis=1,
                        invert_yaxis=True, ylim=(0, 70), yticks=np.arange(0, 50, 10),
                        ylabel='Precipitation (mm/d)', ylabel_loc='top')
    fig.set_layout_axis(subplot=1,
                        ylim=(10, 30), ylabel='Water content at\n80 cm depth (vol.%)',
                        yticks=np.arange(10, 35, 5),
                        text='Compaction top 60 cm', 
                        text_loc=(0.3, 0.85))
    fig.set_layout_axis(subplot=2,
                        ylim=(0, 8), ylabel='Diverted infiltration\n(mm/d)', yticks=np.arange(0, 9))
    fig.set_layout_axis(subplot=3,
                        ylim=(0, 8), yticks=np.arange(0, 9),
                        ylabel='Breakthrough\n(mm/d)',
                        xlim=pd.to_datetime(period),
                        xticks={'loc_maj': mdates.YearLocator(),
                                'for_maj': mdates.DateFormatter('%Y')})
    fig.set_layout_subplot(legend_loc='None')
    fig.set_layout_subplot(subplot=1, legend_loc='upper right')
    data = fig.show()
    return data


def plot_Q_daily_corr_year(db):
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'Q', 'layer': 'CL', 'device': 'tb'},
                      resample={'method': 'sum', 'scale': '1D'},
                      filter_gaps={'gap_max_rel': 0.08})
    ds.add_timeseries({'dtype': 'Q', 'layer': 'CB', 'device': 'tb'},
                      resample={'method': 'sum', 'scale': '1D'},
                      filter_gaps={'gap_max_rel': 0.08})
    data = ds.get_dataframe()
    data.columns = ['Qcl', 'Qcb']
    data = data.sort_index(ascending=False)
    # Plot
    fig, ax = plt.subplots(1, 1, figsize=(9/2.54, 7.5/2.54), dpi=1000, layout='constrained')
    sc = ax.scatter(data.Qcl.values, data.Qcb.values, c=data.index.year,
                    marker='.', s=5,
                    cmap=plt.cm.get_cmap(name="magma_r", lut=17),
                    vmin=2008, vmax=2025)
    cbar = fig.colorbar(ScalarMappable(norm=sc.norm, cmap=sc.cmap), ax=ax,
                        label='Year')
    # Layout
    fig.axes[-1].set_yticks(ticks=np.arange(2009.5, 2024.5),
                            labels=np.arange(2009, 2024))
    fig.axes[-1].set_ylim((2009, 2024))
    fig.axes[-1].tick_params(length=0.1)
    ax.set_ylabel('Breakthrough (mm/d)')
    ax.set_ylim(-0.5, 16)
    ax.set_yticks(np.arange(0, 17))
    ax.set_xlabel('Diverted infiltration (mm/d)')
    ax.set_xlim(-0.5, 7)
    ax.set_xticks(np.arange(0, 7))
    return data


def plot_Q_daily_corr_year_subplots(db):
    # Get data
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'Q', 'layer': 'CL', 'device': 'tb'},
                      resample={'method': 'sum', 'scale': '1D'},
                      filter_gaps={'gap_max_rel': 0.08})
    ds.add_timeseries({'dtype': 'Q', 'layer': 'CB', 'device': 'tb'},
                      resample={'method': 'sum', 'scale': '1D'},
                      filter_gaps={'gap_max_rel': 0.08})
    data = ds.get_dataframe()
    data.columns = ['Qcl', 'Qcb']
    years = data.index.year.unique().values
    # Plot
    fig, ax = plt.subplots(3, 5, figsize=(12, 7), dpi=300, layout='constrained')
    for ii, year in enumerate(years):
        datap = data[data.index.year == year]
        sc = ax[ii % 3, int(ii / 3)].scatter(datap.Qcl.values, datap.Qcb.values,
                                             c=datap.index.year,
                                             marker='.', s=5,
                                             cmap=plt.cm.get_cmap(name="magma_r", lut=17),
                                             vmin=2008, vmax=2025)
        ax[ii % 3, int(ii / 3)].set_ylim(-0.5, 16)
        ax[ii % 3, int(ii / 3)].set_xlim(-0.5, 7)
        ax[ii % 3, int(ii / 3)].set_title(year)
    return data


def plot_watercontent(db, period=None):
    period = period if period else ['2010-05-01', '2014-12-01']
    label_loc = {'up': 'Upslope',
                 'mid': 'Midslope',
                 'down': 'Downslope'}
    # Plot per location
    fig = FigurePlot(db, rows=3, figsize=(19, 13), sharex=True)
    for ii, device in enumerate(['up', 'mid', 'down']):
        for depth in ['5', '15', '40', '80']:
            fig.plot_timeseries({'dtype': 'sm', 'device': device, 'layer': depth+'cm'},
                                period=period, subplot=ii, label=depth + ' cm')
            fig.set_layout_axis(subplot=ii, title=label_loc[device],
                                ylabel='Water content (vol.%)')
    fig.set_layout_axis(subplot=2,
                        xticks={'loc_maj': mdates.MonthLocator(bymonth=[1, 4, 7, 10]),
                                'for_maj': mdates.DateFormatter('%b\n%Y')},
                        xlim=pd.to_datetime(['2010-04', '2015-06']))
    data = fig.show()
    return data


def plot_frequency_distribution_Q_daily(db, bin_max=18,
                                        bin_step=1,
                                        exclude_zeros=False):
    # Get data
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'Q', 'device': 'tb', 'layer': 'CL'})
    ds.add_timeseries({'dtype': 'Q', 'device': 'tb', 'layer': 'CB'})
    ds.set_properties_timeseries(resample={'method': 'sum', 'scale': '1D'},
                                 filter_gaps={'gap_max_rel': 0.08})
    data = ds.get_dataframe()
    # Prepare figure
    fig = FigurePlot(db, figsize=(14, 8))
    labels = ['CL', 'CB']
    bins = np.arange(0, bin_max, bin_step)
    colors = ['C2', 'C3']
    markers = ['.', 'x']
    markersizes = [8, 5]
    hatches = ['////', '----']
    # Plot data
    for i, col in enumerate(data.columns):
        data_ts = data[col].dropna().values
        xdata, freq, freq_cum, load_sum = __calculate_frequencies(data_ts, bins,
                                                                  exclude_zeros)
        fig.plot_data(axis=0, data=pd.Series(index=xdata-0.5, data=freq),
                      bar=True, bar_stack=i, color=colors[i], hatch=hatches[i],
                      label='Outflow '+labels[i])
        fig.plot_data(axis=1, data=pd.Series(index=xdata, data=freq_cum),
                      color=colors[i], linestyle='-', marker=markers[i],
                      markersize=markersizes[i],
                      label='Cumulative frequency flows '+labels[i])
        fig.plot_data(axis=1, data=pd.Series(index=xdata, data=load_sum),
                      color=colors[i], linestyle='--', marker=markers[i],
                      markersize=markersizes[i],
                      label='Cumulative ratio flows to total flow '+labels[i])
    # Set layout
    fig.set_layout_axis(bar_width=0.6*bin_step)
    fig.set_layout_axis(axis=0, ylabel='Number of days',
                        xlabel='Outflow (mm/d)',
                        xticks=np.arange(0, bin_max, 1),
                        ylim=(0, 5050),
                        xlim=(-bin_step, bin_max))
    fig.set_layout_axis(axis=1,
                        ylabel='Cumulative frequency outflow\nCumulative ratio flow to total flow (%)',
                        ylim=(0, 101), yticks=np.arange(0, 110, 10))
    fig.set_layout_axis(grid_linestyle='--')
    fig.set_layout_subplot(legend_loc='center right')
    return fig.show()


def plot_frequency_distribution_Q_daily_subset(db, bin_max=18,
                                               bin_step=0.1,
                                               exclude_zeros=False):
    # Get data
    ds = Dataset(db)
    ds.add_timeseries({'dtype': 'Q', 'device': 'tb', 'layer': 'CL'})
    ds.add_timeseries({'dtype': 'Q', 'device': 'tb', 'layer': 'CB'})
    ds.set_properties_timeseries(resample={'method': 'sum', 'scale': '1D'},
                                 filter_gaps={'gap_max_rel': 0.08})
    data = ds.get_dataframe()
    # Prepare figure
    fig = FigurePlot(db, figsize=(14, 8))
    labels = ['CL', 'CB']
    bins = np.arange(0, bin_max, bin_step)
    colors = ['C2', 'C3']
    markers = ['.', 'x']
    markersizes = [8, 5]
    hatches = ['////', '----']
    # Plot data
    for i, col in enumerate(data.columns):
        data_ts = data[col].dropna().values
        xdata, freq, freq_cum, load_sum = __calculate_frequencies(data_ts, bins,
                                                                  exclude_zeros)
        fig.plot_data(axis=0, data=pd.Series(index=xdata-0.05, data=freq),
                      bar=True, bar_stack=i, color=colors[i], hatch=hatches[i],
                      label='Outflow '+labels[i])
        fig.plot_data(axis=1, data=pd.Series(index=xdata, data=freq_cum),
                      color=colors[i], linestyle='-', marker=markers[i],
                      markersize=markersizes[i],
                      label='Cumulative frequency flows '+labels[i])
        fig.plot_data(axis=1, data=pd.Series(index=xdata, data=load_sum),
                      color=colors[i], linestyle='--', marker=markers[i],
                      markersize=markersizes[i],
                      label='Cumulative ratio flows to total flow '+labels[i])
    # Set layout
    fig.set_layout_axis(bar_width=0.6*bin_step)
    fig.set_layout_axis(axis=0, ylabel='Number of days',
                        xlabel='Outflow (mm/d)',
                        xticks={'ticks': np.arange(0, 1.1, 0.1),
                                'labels': [f'{i:.1f}' for i in np.arange(0, 1.1, 0.1)]},
                        ylim=(0, 5050),
                        xlim=(-0.05, 1))
    fig.set_layout_axis(axis=1,
                        ylabel='Cumulative frequency outflow\nCumulative ratio flow to total flow (%)',
                        ylim=(0, 101), yticks=np.arange(0, 110, 10))
    fig.set_layout_axis(grid_linestyle='--')
    fig.set_layout_subplot(legend_loc=(0.3, 0.4))
    return fig.show()


def __calculate_frequencies(data, bins, exclude_zeros):
    # Get sum of elements inside bins
    bin_data = np.digitize(data, bins=bins, right=True)
    freq = np.bincount(bin_data, minlength=len(bins))
    # Plot frequency discrete and cumulatively
    xdata = bins[1:] if exclude_zeros else bins
    freq = freq[1:] if exclude_zeros else freq
    freq_cum = np.cumsum(freq)
    freq_cum = freq_cum / freq_cum[-1] * 100
    M = sp.sparse.csr_matrix((data, bin_data, np.arange(data.shape[0]+1)),
                             shape=(data.shape[0], len(bins)))
    load_sum = np.asarray(M.sum(axis=0)).cumsum()
    load_sum = load_sum / load_sum[-1] * 100
    load_sum = load_sum[1:] if exclude_zeros else load_sum
    return [xdata, freq, freq_cum, load_sum]
